#include "brad.h"

// Brad Smith
// 5/10/2009

// note: unrolling the row/column/box tests turned out to be much faster (~2/3)
#define UNROLL_LINE_TESTS 1
#define UNROLL_BOX_TEST 1

// note: is was very slightly faster (~80/83) to clear 3 ints instead of 10 chars.
#define UNION_USED_CLEAR 1

// structure storing the status of the numbers 1 to 9 in a row/col/block
#if UNION_USED_CLEAR
	union UsedBlock
	{
		char v[12]; // v[0] is unused, v[11] and v[12] are padding
		int fill[3]; // faster to set 3 ints to 0 than 12 chars.
	};

	inline void empty_used( UsedBlock & u )
	{
		u.fill[0]=0;
		u.fill[1]=0;
		u.fill[2]=0;
	}
#else
	struct UsedBlock
	{
		char v[10]; // v[0] is unused
	};

	static const UsedBlock EMPTY = { { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 } };
	inline void empty_used( UsedBlock & u )
	{
		u = EMPTY;
	}
#endif

// for filling and testing the UsedBlock

static inline bool is_used( int val, UsedBlock & used )
{
	// note: tried "char val" but was much slower, esp. with the unrolled loops (~7/4)
	if ( !val ) return false; // 0 is empty entry
	if ( used.v[val] ) return true;
	used.v[val] = 1;
	return false;
}

// for testing the row/col/block of a proposed solution

static inline bool test_point( const char * const grid, const int index )
{
	const int x = index % 9;
	const int y = index / 9;

	UsedBlock used;
	// note: tried making "used" static, resulted in 3/2 slowdown.

	// test row
	{
		empty_used(used);

		int t = (y * 9);
#		if UNROLL_LINE_TESTS
			if ( is_used( grid[t  ], used ) ) return false;
			if ( is_used( grid[t+1], used ) ) return false;
			if ( is_used( grid[t+2], used ) ) return false;
			if ( is_used( grid[t+3], used ) ) return false;
			if ( is_used( grid[t+4], used ) ) return false;
			if ( is_used( grid[t+5], used ) ) return false;
			if ( is_used( grid[t+6], used ) ) return false;
			if ( is_used( grid[t+7], used ) ) return false;
			if ( is_used( grid[t+8], used ) ) return false;
#		else
			for ( int i=9; i ; --i )
			{
				if ( is_used( grid[t], used ) ) return false;
				++t;
			}
#		endif
	}

	// test col
	{
		empty_used(used);

		int t = x;
#		if UNROLL_LINE_TESTS
			if ( is_used( grid[t   ], used ) ) return false;
			if ( is_used( grid[t+ 9], used ) ) return false;
			if ( is_used( grid[t+18], used ) ) return false;
			if ( is_used( grid[t+27], used ) ) return false;
			if ( is_used( grid[t+36], used ) ) return false;
			if ( is_used( grid[t+45], used ) ) return false;
			if ( is_used( grid[t+54], used ) ) return false;
			if ( is_used( grid[t+63], used ) ) return false;
			if ( is_used( grid[t+72], used ) ) return false;
#		else
			for ( int i=9; i ; --i )
			{
				if ( is_used( grid[t], used ) ) return false;
				t += 9;
			}
#		endif
	}

	// test box
	{
		const int bx = x - (x % 3);
		const int by = y - (y % 3);

		empty_used(used);

#		if UNROLL_BOX_TEST
			int t = bx + (by * 9);
			if ( is_used( grid[t      ], used ) ) return false;
			if ( is_used( grid[t+ 1   ], used ) ) return false;
			if ( is_used( grid[t+ 2   ], used ) ) return false;
			if ( is_used( grid[t   + 9], used ) ) return false;
			if ( is_used( grid[t+ 1+ 9], used ) ) return false;
			if ( is_used( grid[t+ 2+ 9], used ) ) return false;
			if ( is_used( grid[t   +18], used ) ) return false;
			if ( is_used( grid[t+ 1+18], used ) ) return false;
			if ( is_used( grid[t+ 2+18], used ) ) return false;
#		else
			for ( int j=0; j<3; ++j )
			for ( int i=0; i<3; ++i )
			{
				if ( is_used(grid[(i+bx)+((j+by)*9)], used ) ) return false;
			}
#		endif
	}

	return true;
}

// recursive solution finder

static bool ice_climbers_recursive( char * grid, int index )
{
	// finished!
	if ( index >= 81 )
	{
		return true;
	}

	// skip squares already dictated by problem
	if ( grid[index] )
	{
		return ice_climbers_recursive( grid, index + 1 );
	}

	// test the 9 possible values
	for ( char c = 1; c <= 9; ++c )
	{
		// place test value in grid, and see where it goes
		grid[index] = c;
		if ( test_point( grid, index ) )
		{
			if ( ice_climbers_recursive( grid, index + 1 ) )
			{
				return true;
			}
		}
	}

	// there was no solution found in this branch, remove our test value
	grid[index] = 0;
	return false;
}

// finds the next useable empty slot

static inline int next_empty( const char * problem, int i )
{
	while ( i<81 && problem[i] ) ++i;
	return i;
}

static inline int prev_empty( const char * problem, int i )
{
	while ( problem[i] ) --i;
	// note: because there is a guaranteed solution,
	//       there is no danger of falling off the left of the array
	return i;
}

// copies problem into solution
static inline void copy_problem( const char * problem, char * solution )
{
	for ( int i=0; i < 81; ++i )
	{
		solution[i] = problem[i];
	}
}

// entry points
namespace Brad
{
	void ice_climbers( const char * problem, char * solution )
	{
		copy_problem( problem, solution );
		ice_climbers_recursive( solution, 0 );
	}

	void ice_drillers( const char * problem, char * solution )
	{
		copy_problem( problem, solution );
		const int lowest = next_empty(problem,0);
		int i = lowest;
		while ( i < 81 )
		{
			// try next character
			const char try_c = solution[i] + 1;
			if ( try_c > 9 ) // no more solutions when we've reached 10
			{
				solution[i] = 0;
				i = prev_empty(problem,i-1);
			}
			else
			{
				solution[i] = try_c;
				if ( test_point(solution,i) )
				{
					i = next_empty(problem,i+1);
				}
			}
		}
	}
}

// END OF FILE
