#include "generator.h"

#include <cstdlib> // exit
#include <cstdio> // printf

// prints an error message and quits, used to protect against a faulty generator
static void fatal_error( const char * msg )
{
	::printf("Fatal generator error: %s\n",msg);
	exit(-1);
}

// simple random number generator
static unsigned int rand9_seed = 0;

namespace Generator {
	unsigned int rand9()
	{
		// linear congruential generator, numbers used are
		// relatively prime to both 2^32 and 9
		rand9_seed = ( rand9_seed * 5 * 7 * 11 * 13 ) + 17;
		return rand9_seed % 9;
	}
}

// are all numbers in solution between 1 and 9
static bool valid_numbers( const char * solution )
{
	for ( int i=0; i < 81; ++i )
	{
		char value = solution[i];
		if ( value < 1 || value > 9 )
		{
			return false;
		}
	}
	return true;
}

// turns a grid coordinate x,y into an index
static int coord(int x, int y)
{
	return x + (y * 9);
}

// are there no duplicates from 1 to 9 in a row
static bool valid_row( const char * solution, int row )
{
	bool used[9] = {
		false, false, false,
		false, false, false,
		false, false, false };

	for ( int i=0; i < 9; ++i )
	{
		char value = solution[ coord(i,row) ];
		if ( value >= 1 && value <= 9 )
		{
			int index = value-1;
			if ( used[index] ) return false;
			used[index] = true;
		}
	}

	return true;
}

// are there no duplicates from 1 to 9 in a column
static bool valid_col( const char * solution, int col )
{
	bool used[9] = {
		false, false, false,
		false, false, false,
		false, false, false };

	for ( int i=0; i < 9; ++i )
	{
		char value = solution[ coord(col,i) ];
		if ( value >= 1 && value <= 9 )
		{
			int index = value-1;
			if ( used[index] ) return false;
			used[index] = true;
		}
	}

	return true;
}

// are there no duplicates from 1 to 9 in a 3x3 box
static bool valid_box( const char * solution, int x, int y )
{
	bool used[9] = {
		false, false, false,
		false, false, false,
		false, false, false };

	for ( int iy=0; iy < 3; ++iy )
	for ( int ix=0; ix < 3; ++ix )
	{
		char value = solution[ coord(ix+x,iy+y) ];
		if ( value >= 1 && value <= 9 )
		{
			int index = value-1;
			if ( used[index] ) return false;
			used[index] = true;
		}
	}

	return true;
}

// checks all rows, columns, and 3x3 boxes for validity
static bool valid_rcb( const char * solution )
{
	for ( int i=0; i < 9; ++i )
	{
		if ( !valid_row( solution, i ) ) return false;
		if ( !valid_col( solution, i ) ) return false;
	}

	for ( int y=0; y<9; y+=3 )
	for ( int x=0; x<9; x+=3 )
	{
		if ( !valid_box( solution, x, y ) ) return false;
	}

	return true;
}

// are all nonempty members of the problem present in the solution
static bool valid_match( const char * problem, const char * solution )
{
	for ( int i=0; i < 81; ++i )
	{
		if ( problem[i] != 0 && problem[i] != solution[i] )
		{
			return false;
		}
	}
	return true;
}

// recursive function that picks a random empty entry to fill
// and attempts to solve it. returns true if a complete solution has
// been found, false if no solutions could be found from this branch.
// pass the problem, a list of 81 unique ints (0-80) indicating the
// order in which to fill in squares, and an index into that order
static bool generate_recursive( char * problem, int index )
{
	// we must be done
	if ( index >= 81 ) return true;

	// try all 9 numbers in this slot, and recurse from there
	int start = Generator::rand9(); // pick a random number to start with
	for ( int i = 0; i < 9; ++i )
	{
		char c = char(((start+i)%9)+1);
		problem[index] = c;
		if ( valid_rcb( problem ) )
		{
			if ( generate_recursive( problem, index+1 ) )
			{
				return true;
			}
		}
	}
	// no solution was found, return the problem to its previous state
	problem[index] = 0;
	return false;
}

// prints one row of the sudoku, no newline
static void print_row( const char * row )
{
	for ( int col = 0; col < 9; ++col )
	{
		char value = row[ col ];

		if ( value >= 1 && value <= 9 ) ::printf("%1d", value);
		else ::printf(".");

		if ( col != 8 ) ::printf(" ");
	}
}

namespace Generator
{
	// seeds the random number generator
	void seed_generator( unsigned int seed )
	{
		rand9_seed = seed;
	}

	// generates a new problem
	void generate( char * problem )
	{
		// wipe out the problem
		for ( int i=0; i < 81; ++i )
		{
			problem[i] = 0;
		}

		// recursively generate a valid solution, element by element
		generate_recursive( problem, 0 );

		// verify solution
		if ( !solved( problem, problem ) ) fatal_error("Created invalid problem.");

		//::printf("SOLUTION:\n");
		//print(problem);

		// zero a random number of entries to create the problem from the solution
		for ( int i=0; i<81; ++i )
		{
			if ( 0 == (rand9() % 2) ) // 5:4 odds
			{
				problem[i] = 0;
			}
		}

		//::printf("PROBLEM:\n");
		//print(problem);
	}

	// tests for a valid solution
	bool solved( const char * problem, const char * solution )
	{
		if ( !valid_match( problem, solution ) ) return false;
		if ( !valid_numbers( solution ) ) return false;
		if ( !valid_rcb( solution ) ) return false;
		return true;
	}

	// prints a sudoku
	void print( const char * sudoku )
	{
		for ( int row = 0; row < 9; ++row )
		{
			print_row( sudoku + (row * 9) );
			::printf("\n");
		}
	}

	// prints two sudoku side by side for comparison
	void print2( const char * left, const char * right )
	{
		for ( int row = 0; row < 9; ++row )
		{
			print_row( left + (row * 9) );
			::printf("       ");
			print_row( right + (row * 9) );
			::printf("\n");
		}

	}

}
// END OF FILE
