#include <cstdio> // for printf
#include <windows.h> // for QueryPerformanceCounter/QueryPerformanceFrequency
#include "generator.h"

const int NUM_PROBLEMS = 30;
// define this as 1 to print all solutions to the console
#define SHOW_SOLUTIONS 0

// define this as 1 to maximize thread priority
#define TIME_CRITICAL 1

// solvers included by header
#include "example.h"
#include "brad.h"
#include "BetterThan10.h"
#include "brock.h"
#include "steve_weatherly.h"


// array of solver programs to be tested
struct SolverTest
{
	const char * name;
	void (*solver)( const char * problem, char * solution );
};

static const SolverTest tests[] =
{
	{ "Example::bad_answer", Example::bad_answer },
	{ "Brad::ice_climbers", Brad::ice_climbers },
	{ "Brad::ice_drillers", Brad::ice_drillers },
	{ "BetterThan10::SolveSimple", BetterThan10::SolveSimple },
	{ "BetterThan10::SolveOrdered", BetterThan10::SolveOrdered },
	{ "BetterThan10::SolveMangled", BetterThan10::SolveMangled },
	{ "BetterThan10::SolveCheater1", BetterThan10::SolveCheater1 },
	{ "BetterThan10::SolveCheater2", BetterThan10::SolveCheater2 },
	{ "Brock::find_solution", Brock::find_solution },
	{ "SteveW::EvolveSolution", SteveW::EvolveSolution },
};

const int NUM_TESTS = sizeof(tests)/sizeof(tests[0]);

// Win32 timer

LARGE_INTEGER start_time;

void start_timer()
{
	::QueryPerformanceCounter(&start_time);
}

double end_timer()
{
	LARGE_INTEGER end_time;
	::QueryPerformanceCounter(&end_time);

	LARGE_INTEGER freq;
	::QueryPerformanceFrequency(&freq);

	LONGLONG diff = end_time.QuadPart - start_time.QuadPart;
	
	return double(diff) / double(freq.QuadPart);
}

// main test

int main(int argc, char ** argv)
{
#	if TIME_CRITICAL
		if ( !::SetThreadPriority( GetCurrentThread(), THREAD_PRIORITY_TIME_CRITICAL ) )
		{
			::printf("Unable to set thread priority.\n");
			return -1;
		}
#	endif TIME_CRITICAL

	char * problems = new char[ 81 * NUM_PROBLEMS ];
	char * solutions = new char[ 81 * NUM_PROBLEMS ];

	start_timer();
	unsigned int seed = start_time.LowPart; // generate a seed from the timer
	Generator::seed_generator(seed);
	::printf("Generating %d problems from seed: %u\n", NUM_PROBLEMS, seed );

	for ( int i=0; i < NUM_PROBLEMS; ++i )
	{
		Generator::generate( problems + (81 * i) );
	}
	::printf("Generation took %12.6f seconds.\n", end_timer() );

	::printf("Performing %d tests...\n",NUM_TESTS);
	// create random test order
	int test_order[NUM_TESTS];
	for ( int i=0; i < NUM_TESTS; ++i )
	{
		test_order[i] = i;
	}
	::srand(seed);
	for ( int i=0; i < NUM_TESTS * 2; ++i )
	{
		int a = Generator::rand9() + ( Generator::rand9() * 9 );
		int b = Generator::rand9() + ( Generator::rand9() * 9 );
		a = a % NUM_TESTS;
		b = b % NUM_TESTS;
		int swap = test_order[a];
		test_order[a] = test_order[b];
		test_order[b] = swap;
	}

	for ( int t=0; t < NUM_TESTS; ++t )
	{
		// wipe previous solutions
		for ( int i=0; i < (81 * NUM_PROBLEMS); ++i )
		{
			solutions[i] = 0;
		}

		char * problem = problems;
		char * solution = solutions;

		// timed test of all problems

		int ot = test_order[t];

		start_timer();

		for ( int p=0; p < NUM_PROBLEMS; ++p )
		{
			tests[ot].solver( problem, solution );
			problem += 81;
			solution += 81;
		}

		double duration = end_timer();

		// verify solutions
		problem = problems;
		solution = solutions;
		int solved = 0;
		for ( int p=0; p < NUM_PROBLEMS; ++p )
		{
			bool solved_this = Generator::solved( problem, solution );
			if ( solved_this ) ++solved;

#if SHOW_SOLUTIONS
			::printf("\n");
			if ( solved_this )
				::printf("SOLVED %d\n",p);
			else
				::printf("FAILED %d\n",p);
			Generator::print2(problem, solution);
#endif

			problem += 81;
			solution += 81;
		}

		// print stats
		::printf("%4d: %40s solved %7d/%7d in %12.6f seconds.\n",
			t, tests[ot].name, solved, NUM_PROBLEMS, duration );
	}
	::printf("Done.\n");

	return 0;
}

// END OF FILE
