#include "steve_weatherly.h"
//#include <cstdlib>
//#include <cstdio>

namespace SteveW
{
	static const int GENE_LENGTH = 81;

	////// Algorithm Setting ////////
	static const int POOL_SIZE = 100;
	static const int MUTATION_CHANCE = 10;//% chance out of 100
	static const int TOURNAMENT_POOL_SIZE = 75;
	static const int PERFECT_FITNESS = 243;
	static const int EXTINCTION_COUNT = 4000;

	//////// Structs //////////
	struct Gene
	{
		Gene() : fitness(-1){}

		int chromosome[GENE_LENGTH];
		int fitness;
	};

	/////Pre-Declarations//////
	inline int coord(int x, int y);
	inline int GetRowScore( const int *solution, int row );
	inline int GetColScore( const int *solution, int col );
	inline int GetBoxScore( const int *solution, int x, int y );	
	inline bool ValueExists(int testVal, int *intList, int size);

	inline void CreateNextGeneration();
	inline void PickParents(int *parentAOut, int *parentBOut);
	inline Gene BreedGenes(const Gene &parentA, const Gene &parentB);
	inline int GetSolutionIndex();
	inline void SetFitness(Gene &gene);
	inline int rand();
	
	//inline int rand(const int max);

	


	//// Local Variables ////////
	static Gene GenePool[POOL_SIZE];	
	static int  Problem[GENE_LENGTH];
	static int  Seed;

	///////////////// ENTRY POINT ////////////////////////
	void EvolveSolution( const char * problem, char * solution )
	{		
		Seed = 0;
		//Copy off an easy to use version
		for(int i=0;i<GENE_LENGTH;++i)
		{
			Problem[i] = int(problem[i]);
			Seed += Problem[i];
		}

		
		//Randomly fill gene pool
		for(int i=0;i<POOL_SIZE;++i)
		{
			for(int c=0;c<GENE_LENGTH;++c)
			{
				GenePool[i].chromosome[c] = 1 + (rand() % 9);
			}			
		}

		int solIndex = -1;
		int generationCounter = 0;
		//Evolve a solution or die trying
		do 
		{
			//printf("== Generation %d Being Bred\n", generationCounter);
			CreateNextGeneration();
			solIndex = GetSolutionIndex();
			generationCounter++;
		} while (solIndex < 0 && generationCounter < EXTINCTION_COUNT);
		
		if (generationCounter >= EXTINCTION_COUNT)
		{
			solIndex = 0;
		}

		for(int i=0;i<GENE_LENGTH;++i)
		{
			solution[i] = char(GenePool[solIndex].chromosome[i]);
		}

	}

	////////////// GENETIC ALGORITHM HELPER FUNCTIONS /////////////////


	int GetSolutionIndex()
	{
		//int best = 0;
		for(int i=0;i<POOL_SIZE;++i)
		{
			if (GenePool[i].fitness < 0)
			{				
				SetFitness(GenePool[i]);
			}

//			if (GenePool[i].fitness > best)
//				best = GenePool[i].fitness;

			if (GenePool[i].fitness >= PERFECT_FITNESS)
				return i;
		}		
		//printf("Best: %d\n", best);
		
		return -1;
	}

	void CreateNextGeneration()
	{
		int indexA, indexB;
		PickParents(&indexA, &indexB);
		Gene parentA = GenePool[indexA];
		Gene parentB = GenePool[indexB];

		for(int i=0;i<POOL_SIZE;++i)
		{			
			GenePool[i] = BreedGenes(parentA, parentB);
		}
	}

	Gene BreedGenes(const Gene &parentA, const Gene &parentB)
	{
		Gene child;
		
		//Find crossover points for breeding
		int start = rand() % GENE_LENGTH;
		int stop = rand() % GENE_LENGTH;
		if (start > stop)
		{
			int temp = start;
			start = stop;
			stop = temp;
		}

		for(int i=0;i<GENE_LENGTH;++i)
		{
			if (Problem[i] != 0)//The problem itself has highest priority
				child.chromosome[i] = Problem[i];
			else if (rand() % 100 <= MUTATION_CHANCE)//mutate
				child.chromosome[i] = 1 + (rand() % 9);
			else if (i > start && i < stop)
				child.chromosome[i] = parentB.chromosome[i];//parent b
			else
				child.chromosome[i] = parentA.chromosome[i];//parent a				
		}
		return child;
	}
	
	void SetFitness(Gene &gene)
	{
		gene.fitness = 0;
		for(int i=0;i<9;++i)
		{
			gene.fitness += GetRowScore(gene.chromosome, i);			

			gene.fitness += GetColScore(gene.chromosome, i);			
		}

		for(int y=0;y<9;y+=3)
		{
			for(int x=0;x<9;x+=3)
			{
				gene.fitness += GetBoxScore(gene.chromosome, x, y);					
			}
		}
	}

	void PickParents(int *parentAOut, int *parentBOut)
	{
		int tPool[TOURNAMENT_POOL_SIZE];
		//Find unique candidates for tournament selection
		for(int i=0;i<TOURNAMENT_POOL_SIZE;++i)
		{
			do 
			{
				tPool[i] = rand() % POOL_SIZE;
			} while (i > 0 && ValueExists(tPool[i], tPool, i));			
		}

		//Find the best two candidates
		int bestI = -1;
		int nextBestI = -1;
		int bestScore = -1;
		int nextBestScore = -1;

		for(int i=0;i<TOURNAMENT_POOL_SIZE;++i)
		{
			if (GenePool[tPool[i]].fitness < 0)
				SetFitness(GenePool[tPool[i]]);

			if (GenePool[tPool[i]].fitness > bestScore)
			{
				nextBestI = bestI;
				nextBestScore = bestScore;
				bestI = i;
				bestScore = GenePool[tPool[i]].fitness;
			}
			else if (GenePool[tPool[i]].fitness > nextBestScore)
			{
				nextBestI = i;
				nextBestScore = GenePool[tPool[i]].fitness;				
			}
		}
		*parentAOut = tPool[bestI];
		*parentBOut = tPool[nextBestI];

	}	

	//////////// Other Helper Functions ////////////////////

 	int rand()
 	{
 		static long a = Seed;  //seed value
 
 		a = (a * 32719 + 3) % 32749;
 		return a;
 	}

	bool ValueExists(int testVal, int *intList, int size)
	{
		for(int i=0;i<size;++i)
		{
			if (intList[i] == testVal)
				return true;
		}
		return false;
	}

	// turns a grid coordinate x,y into an index
	int coord(int x, int y)
	{
		return x + (y * 9);
	}

	// are there no duplicates from 1 to 9 in a row
	int GetRowScore( const int *solution, int row )
	{
		int score = 0;

		bool used[9] = {
			false, false, false,
			false, false, false,
			false, false, false };

			for ( int i=0; i < 9; ++i )
			{
				int value = solution[ coord(i,row) ];
				if ( value >= 1 && value <= 9 )
				{
					int index = value-1;
					if ( !used[index] )
						score++;

					used[index] = true;
				}
			}

			return score;
	}

	// are there no duplicates from 1 to 9 in a column
	int GetColScore( const int *solution, int col )
	{
		int score=0;
		bool used[9] = {
			false, false, false,
			false, false, false,
			false, false, false };

			for ( int i=0; i < 9; ++i )
			{
				int value = solution[ coord(col,i) ];
				if ( value >= 1 && value <= 9 )
				{
					int index = value-1;
					if ( !used[index] )
						score++;
					used[index] = true;
				}
			}

			return score;
	}

	// are there no duplicates from 1 to 9 in a 3x3 box
	int GetBoxScore( const int *solution, int x, int y )
	{
		int score=0;
		bool used[9] = {
			false, false, false,
			false, false, false,
			false, false, false };

			for ( int iy=0; iy < 3; ++iy )
				for ( int ix=0; ix < 3; ++ix )
				{
					int value = solution[ coord(ix+x,iy+y) ];
					if ( value >= 1 && value <= 9 )
					{
						int index = value-1;
						if ( !used[index] )
							score++;
						used[index] = true;
					}
				}

				return score;
	}	
	
}
